/***************************************************************************
 *   Copyright (C) 2005 Nicolas Hadacek <hadacek@kde.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#ifndef PROCESS_H
#define PROCESS_H

#include <qstringlist.h>
#include <qobject.h>
#if QT_VERSION<0x040000
class QProcess;
#  define Q3Process QProcess
#else
class Q3Process;
#endif

#include "global.h"
namespace PURL { class Directory; }

namespace Process
{
//----------------------------------------------------------------------------
enum State { Stopped, StartFailed, Running, Exited, Timedout };
enum RunAction { NoRunAction = 0, Start = 1 };
Q_DECLARE_FLAGS(RunActions, RunAction)
Q_DECLARE_OPERATORS_FOR_FLAGS(RunActions)
class Base;
extern State runSynchronously(Base &process, RunActions actions, uint timeout); // in ms (0 == no timeout)

//----------------------------------------------------------------------------
class Base : public QObject
{
Q_OBJECT
public:
  Base(QObject *parent, const char *name);
  virtual ~Base();
  void setup(const QString &executable, const QStringList &options, bool withWine);
  QStringList arguments() const { return _arguments; }
  QString prettyCommand() const { return arguments().join(" "); }
  void setWorkingDirectory(const PURL::Directory &dir);
  void setUseShell(bool useShell);
  virtual bool start(uint timeout); // in ms (0 == no timeout)
  void writeToStdin(const QString &s);
  bool signal(int n);
  bool isRunning() const;
  State state() const { return _state; }
  int exitCode() const;

signals:
  void requestSynchronousStop();
  void done(int code);
  void timeout();
  void stdoutDataReceived();
  void stderrDataReceived();

protected slots:
  void exitedSlot();
  void timeoutSlot();
  void readyReadStdoutSlot();
  void readyReadStderrSlot();

  friend State runSynchronously(Base &, RunActions, uint);

protected:
  Q3Process  *_process;
  QStringList _arguments, _environment;
  bool        _useShell;
  State       _state;
  QTimer     *_timer;
  QString    _stdout, _stderr;

  virtual void receivedStdout(const QString &s) = 0;
  virtual void receivedStderr(const QString &s) = 0;
  static bool isFilteredLine(const QString &line);
};

//----------------------------------------------------------------------------
class StringOutput : public Base
{
Q_OBJECT
public:
  StringOutput(QObject *parent = 0, const char *name = 0) : Base(parent, name) {}
  QString sout() const { return _stdout; }
  QString serr() const { return _stderr; }

private:
  virtual void receivedStdout(const QString &s) { _stdout += s; }
  virtual void receivedStderr(const QString &s) { _stderr += s; }
};

//----------------------------------------------------------------------------
class LineBase : public Base
{
Q_OBJECT
public:
  LineBase(QObject *parent = 0, const char *name = 0) : Base(parent, name) {}

private:
  virtual void addStdoutLine(const QString &line) = 0;
  virtual void addStderrLine(const QString &line) = 0;
  virtual void receivedStdout(const QString &s);
  virtual void receivedStderr(const QString &s);
};

//----------------------------------------------------------------------------
class LineOutput : public LineBase
{
Q_OBJECT
public:
  LineOutput(QObject *parent = 0, const char *name = 0) : LineBase(parent, name) {}
  virtual bool start(uint timeout);
  QStringList sout() const { return _stdoutLines; }
  QStringList serr() const { return _stderrLines; }

protected:
  QStringList _stdoutLines, _stderrLines;

  virtual void addStdoutLine(const QString &line) { _stdoutLines += line; }
  virtual void addStderrLine(const QString &line) { _stderrLines += line; }
};

//----------------------------------------------------------------------------
class LineSignal : public LineBase
{
Q_OBJECT
public:
  LineSignal(QObject *parent = 0, const char *name = 0) : LineBase(parent, name) {}

signals:
  void logStdoutLine(const QString &line);
  void logStderrLine(const QString &line);

private:
  virtual void addStdoutLine(const QString &line) { emit logStdoutLine(line); }
  virtual void addStderrLine(const QString &line) { emit logStderrLine(line); }
};

} // namespace

#endif
